﻿#pragma once
#include <QPixmap>
#include <QPainter>
#include <QPointF>
#include <QColor>
#include <QRectF>
#include <QImage>
#include <QFontMetrics>
#include <QSize>
#include <QPainterPath>
#include <QMatrix>
#include <QTime>
#include <QResizeEvent>
#include "MathExt.h"
#include "track.h"
#include "HeightData.h"
#include "osm.h"
#include <QtWidgets/QMessageBox>

//! generating maps
/*! generating maps using SRTM3 und btp (imported OSM data) data in mercator
 *  projection. The resulting image of the current map is stored in cmap.
 *  The drawing routine is located in draw_map() function.
*/
class MAP : public QWidget{
        Q_OBJECT
public:
    //! create with SRTM3 class and OSM-data manager
    MAP(HeightData *h, OSM* o);
    MAP(QWidget *parent);
    void set_data(HeightData *h, OSM* o);
	~MAP(void);
    //! assigns new settings
    void settings(MapSettings* MS);
    //! adapts size, recalculate borders from desired borders
    void resize(int width, int height);
    //! sets desired borders, recalculate bordes due to window size
    void set_borders(double latmin0, double latmax0,
                     double lonmin0, double lonmax0);
    //! desired borders to real borders
    /*!
      Desired borders may not represent the rectangle defined width an height.
      New borders will be set completely laying inside the old borders. If a to
      high area was specified, the lattitude range will be reduce, for a to
      wide area the longitude range will be shrinked.
      The method also sets tempory memory.
    */
    void calc_borders();
    //! draws the relief
    void SRTM3toMAP(short getcmap);
    //! draws Open Street Map data
    void OSMtoMAP(short getcmap, short drawroads, short drawnature , short drawpoints);
    //! draws STRONG calculation network
    void STRONGlayer_to_Map(QPainter* p,STRONGlayer* SL);
    //! draws climbs
    void climblayer_to_map(climblayer* cl);
    //! projects best climbs to the borders of the picture
    void add_BestClimb_Bar(QPainter* p,climblayer* cl);
    //! universal high quality drawing
    /*! main mathod of the MAP class, including OSM, SRTM3, STRONGlayer,
     *  CLimblayer. First renders relief from HeightData, then filling
     *  face data of OSM (wood, water) with edge flag algorithm. Then from
     *  least important to most important Map content is drawn onto map. If
     *  a climb layer is give the climb segments and labels are included in
     *  this process. If a STRONGlayer is given, the roads will be painted due
     *  to its STRONGindex
     */
    void draw_map(STRONGlayer* SL, climblayer* cl, Track* ct, int fastmode);
    //! initializes map with size and borders
    void init_map(int width, int height);
    //! save map as picture to hard disk
    void save_to_png(QString filename);
    //! return pixmap of the last generated map
    QPixmap* get_Map();
    //! returns px color
    QColor color(double lon, double lat);
    //! integrates colors alog path
    void color_integral(double lon, double lat, double lastlon,
                        double lastlat, double* samplecount,
                        double* red, double* green, double* blue,QImage* Img);
    //! accelerated function using precalculated Image
    QColor color(double lon, double lat, QImage* Img);
    //! accelerated function using precalculated Image
    QColor color_xy(double x, double y, QImage* Img);
    //! drawing commends from outside
    void draw_line(QPainter* p, KOOlist* k, int kc);
    void draw_line(QPainter* p, KOO** k, int count);
    void draw_line(QPainter* p, double** k, int count);
    void draw_line(QPainter* p, Track* ct);
    void draw_elements(Track* ct, KOO* kold, KOO* knew, Way* selected_way);
    void show_elements(Way* w, int X, int Y, QBrush b);
    void show_zoombox(int x, int xold, int y, int yold);
    void show_translated(int x, int xold, int y, int yold);

    int Draw_lighted_text(QPainter* p, KOO *k, QString* w,
                          QColor* lightcolor, QColor* textcolor);
    //! return real borders
    double getminlat();
    double getminlon();
    double getmaxlat();
    double getmaxlon();
    int getwidth();
    int getheight();
    //! map coordinate information
    double y_to_lat(int y);             //! map coordinates
    double x_to_lon(int x);             //! map coordinates
    double Y_to_lat(int Y);             //! Widget coordinates
    double X_to_lon(int X);             //! Widget coordinates
    double lat_to_y(double lat);        //! map coordinates
    double lon_to_x(double lon);        //! map coordinates
    double xy_to_height(int x, int y);  //! map coordinates
    double XY_to_height(int X, int Y);  //! Widget coordinates
    float zoomsize(float size);         /*!< normalizes font size to zoom*/
    double scale();						/*!< km pro Pixel*/
    float  get_mperlon();
    float  get_mperlat();
    OSM* o;
    void scale(double zoomfactor,double dlonfactor, double dlatfactor);
public slots:
    void adapt_map_size_to_window();
signals:
    void position(double lat, double lon, double height);
    void clicked(double lat, double lon, bool ctrl);
    void keypressed(QKeyEvent *event);
    void new_rawmap();

private:
    int x, y, xold, yold;
    /*void mousePressEvent(QMouseEvent *event);
    void mouseReleaseEvent(QMouseEvent *event);
    void mouseMoveEvent(QMouseEvent *event);
    void wheelEvent(QWheelEvent * event );
    void keyPressEvent(QKeyEvent *);*/
    //! transform widget coordinate(X,Y) into map coordinates
    int X_to_x(int X);
    int Y_to_y(int Y);
    int y_to_Y(int y);
    int x_to_X(int x);
    int paintmode;
    QRectF zoomrect;
    void init();
    void new_map_memory();
    //! specified picture format
    int width, height, new_width, new_height;
    //! specify, whether size adaptation needs to be performed
    bool was_resized;
    //! desired area with ordinary coordinates
    double  minlat0,                    /*!< desired minimum lattitude*/
            maxlat0,                    /*!< desired maximum lattitude*/
            minlon0,                    /*!< desired minimum longitude*/
            maxlon0;                    /*!< desired minimum longitude*/
    //! smaller area inside desired conform width picture scales
    double  minlat,                     /*!< realized minimum lattitude*/
            maxlat,                     /*!< realized maximum lattitude*/
            minlon,                     /*!< realized minimum longitude*/
            maxlon;                     /*!< realized minimum longitude*/
    //! marcator projection scaled lattitude
    double mmaxlat,
           mminlat;

    HeightData* hd;
    QPixmap* cmap, *cmapf, *cmapff;
    MapSettings *MS,MS0;
    //! transform heighdata with sun settings into colors
    void add_DEM();
    //! adds contours to relief
    void add_contour();
    //! function to height to color operation
    void htocolorfast(unsigned char* cd,float h,float dxf,float dyf);


    /*! coordinate transformation between picture coordinates und geografic
    coordinates*/
    double ytolat(int y);
    double xtolon(int x);
    float lattoy(double lat);
    float lontox(double lon);

    //! help data to render picture
    QPointF HP[28];
    double** hc;						/*!< height at px position*/
    double** dhdx;						/*!< gradient in x direction*/
    double** dhdy;						/*!< gradient in y direction*/
    short** h;							/*!< SRTM3 grid, used by SRTM3toMap*/
    unsigned char* cd;					/*!< color data, every px consists of
                                        three bytes, used by all rendering*/
    short lineos;						/*!< line offset to meet byte count
                                         divisible by four for a picture line*/
    short wxxold;						/*!< SRTM3 data cells in x direction*/
    short wyyold;						/*!< SRTM3 data cells iny direction */
    int lowcontour;                     //!< small contour step
    int highcontour;                    //!< big contour step
    float mperlat;                      //!< meter per SRTM3 cell vertical
    float mperlon;                      //!< meter per SRTM3 cell horizontal
    float pxperlat;                     //!< px per SRTM3 cell vertical
    float pxperlon;                     //!< px per SRTM3 cell horizontal
    float zoom;							//!< meter per px
    short drawcontour;
    //! color functions
    QColor colorgradient(float x);
    QColor colorgradientSL(float x);
    QColor colorgradientSchading(float x);
    void change_intensity(int x, int y, //!< changes grayvalue(x,y) by dI
                          double dI);
    void mult_color(unsigned char* cd, QColor* c);
    //! OSM drawing
    void linetomap(OSM* o, QPainter *p, lines** ll,QColor QColor,
                   float linewidth, float minlinewidth);
    void pointtomap(OSM* o, QPainter *p, points** pp,QColor QColor,float size);
    void peaktomap(OSM* o, QPainter *p, points** pp, float min, float max,
                   QColor QColor,float size);
    void waternametomap(OSM* o, QPainter *p, lines** pp,QColor QColor,float size);
    void streetnametomap(OSM* o, QPainter *p, lines** pp,QColor QColor,float size);
    void rivernametomap(OSM* o, QPainter *p, lines** pp,QColor QColor,float size);
    void add_scale(QPainter *p);
    //! draws text with lighted background
    /*! draws text with lighted background; speedtest results for different
        background renderings:
        drawText() 28 times surround transparent: 1.4 ms
        drawPath() 28 times surrounded transparent: 71.6 ms,
        drawPath() 4 times surrounded transparent: 15.1 ms,
        drawPath() with 1 penwidthes: 10.6 ms
        drawPath() with 3 penwidthes: 25.2 ms
    */
    void Draw_lighted_text(QPainter *p, QString* w, QPointF P,
                           QColor* lightcolor, QColor* textcolor);
    void Draw_lighted_Path(QPainter* g, QPainterPath* p,
                           QBrush* lightbrush,QBrush* textbrush);
    int Draw_aligned_text(QPainter* g,QPointF* path,int pointcount, QString*w,
                          QPointF P,QBrush* lightbrush, QBrush* textbrush);


    /*! templated marking occupied map area
        0: unocuppied
        1: wood only
       10: water only
       11: wood and water
    */
    char** t;
    //! clears tamplate
    void reset_Template();
    //! edge flag algorithm to draw polyedges
    void linestoTemplate(lines* f,char val);
    //! fills areas marked by edge flag algorithm
    void facetoTemplate(lines** ff,char val);
    void TemplatetoColorData();
    //! edge flag algorithm marking the px(x,y), used by linestoTemplate
    void set(int x, int y,char val);
    //! check R on the picture if free
    short check_TemplateSpace(QRectF R);
    //! marks R on the picture as occupied
    void set_TemplateSpace(QRectF R);
    //! check R for l-value on the picture if free
    short check_TemplateSpace(QRectF R, char l);
    //! marks R as l-value on the picture as occupied
    void set_TemplateSpace(QRectF R, char l);
    //! quantifies l-content in r-region
    int IntSec(QRectF R, char l);
    //! climb layer method: SRTM3 DEM plus shading of climbs
    void SRTM3ClimbtoMAP(climblayer* cl);
    //! climb layer method: shading called from SRTM3ClimbtoMAP()
    void climblayer_schading(climblayer* cl);
    //! climb layer method: draw tracks
    void ClimbTracks(QPainter*p,climblayer* cl);
    //! climb layer method: draw labels
    void ClimbLabels(QPainter*p,climblayer* cl, int fastmode);
protected:
    void paintEvent(QPaintEvent *);
    void resizeEvent(QResizeEvent *event);
};

