﻿#pragma once

#include "DataTyps.h"
#include "MathExt.h"
#include "HeightData.h"
#include "OSM.h"
#include "MAP.h"
#include "fibunacci.h"
#include "track.h"
#include "QApplication"
#include "QClipboard"
#include "tsp.h"
#include <windows.h>


//! class to perform routing routines on osm data sets
/*! this class provides routing routines to work on an Neighbour network.
 *  STRONGcalc does this as well, but in a seperate thread. Here, Dijkstra
 *  algorithms are implemented.
 *  Additionally this class handels the STRONGlayer buisness, i.g. creating
 *  map layers from finished STRONGcalc calculation.
*/
class Routing{
	public:
        Routing(OSM* o);
		~Routing(void);

        Track* airline(HeightData* h, KOO* s, KOO* e);
        //! this is famous Dijkstra shortes path algothim
        /*! calculates shortes path on an osm dataset using Dijkstra algorithm.
         *  a FibunacciHeap priority queue is used for best performance. When
         *  algorithm finishes there will we a network of STRONG references from
         *  which the shortes paths to the startpoint s can be traced back by
         *  iterating over the KOO->cd->s[Si]->from reference.
         *  The algorithm respects the road restrictions min and max describing
         *  allowed road types. If no way to the end point e is found zero will
         *  be returned, otherwise 1.
         */
        short Dijkstra(KOO* s, KOO* e, char max, char min, OSM* o, int Si,
                       float (*attr)(Neighbour* n));
        //! solves the salesman problem
        /*! based on the array t0 of tc tracks all the best order of tracks is
         *  searched to find the shortes total path reaching all tracks. If no
         *  start- or en points are specified the total path starts and/or ends
         *  at any optimal track t[i]
         */
        short salesman(KOO* s, KOO* e, char max, char min, OSM* o,
                       Track** t0, int tc, Track** result, bool *unreachable);
        //! collects a Neighbour list from STRONG network
        /*! tries to find a path from e to s and collects a list of Neighbour
         *  relations which is used for that purpose. This list can be used to
         *  create Track objects.
         */
        void extract_neighbourlist(STRONG* s, STRONG* e,
                                   Neighbour ***nl,int* nc);
        //! resets the STRONG network to begin calculation
        void resetSTRONG(KOO* k, int STRONGindex);
        //! resets STRONG network of a STRONGcalc to normal Dijkstra state
        /*! normal state of STRONG network consistsof crossdata with STRONG
         *  arrays of length two. Dijkstra will usually use only one of them
         *  but when modifying a track both ar used. In the State for
         *  STRONGcalc these arrays may contain several 100 elements depending
         *  on the setup of the STRONGcalc
         */
        void finish_STRONG();
        //! creates STRONGlayer from a unfinalized STRONG network
		STRONGlayer* STRONGlayer_create(STRONGlayer* SL,int N);
        //! creates STRONGlayer from a unfinalized STRONG network
        STRONGlayer* STRONGlayer_create(int N);
        //! loades STRONGlayer from file
		STRONGlayer* STRONGlayer_load(char* filename);
        //! saves STRONGlayer to file
		void STRONGlayer_save(char* filename, STRONGlayer *SL);
        static float get_hm(Neighbour* n);  //!< returns hm-value, for Dijkstra
        static float get_d(Neighbour* n);   //!< returns d-value, for Dijkstra
        static float get_t(Neighbour* n);   //!< returns t-value, for Dijkstra
        Track* shortest_track(KOO* s, KOO* e, char max, char min, OSM* o);
        double free_RAM_Bytes();
	private:
		OSM* o;
        //! body of finish_STRONG();
        void setSTRONG(KOO* k, KOO* s, KOO* e, float radius, int n);

        //! sets Neighbour distance attribute d to zero
		void STRONGlayer_reset_d(KOO* k);
        //! reconstructs Neighbour distance attribute d from Way data
		void STRONGlayer_reconstruct_d(KOO* k);
        //! counts total Neighbour count on osm data set
        void STRONGlayer_count_neighbours(KOO* k, int *nbc);
        //! increment Neighbour attribute d by its usage rate on STRONG network
		void STRONGlayer_mark_neighbours(KOO* k, int si);
        //! copy Neighbour attribut d to STRONGlayer data
		void STRONGlayer_fill_neighbours(KOO* k, STRONGlayerele* data, 
                                          int *nbc);
        //! quicksort neighbours by its usage rate
		void STRONGlayer_sort(STRONGlayer* SL, long s, long e);
        //!@{
        //! grid to decide whether a cross can still be used to mark neighbours
        /*! to homogeniously probe the STRONG network to reconstruct
         *  the STRONGlayer it is necessary to check that probe points, i.g.
         *  cross are equally distant to each other. Therefor a net with
         *  0.5x0.5 km cells is created, first cross belonging to this cell
         *  will probe the STRONG network, all following cross belonging to
         *  this cell will be ignored thus avoiding overjudging of regions with
         *  high cross density, e.g. big citys
         */
		char** SLgrid;
		int nlat, nlon;
		double SLlonmax,SLlatmax,SLlonmin,SLlatmin,dlat,dlon;
        void STRONGlayer_grid_new();
        void STRONGlayer_grid_reset();
        short STRONGlayer_grid_free(KOO* k);
        void STRONGlayer_grid_delete();
        void deactivate_neighbours(Neighbour** n, int nc);
        void restore_neighbours(Neighbour** n, int nc);
        //!@}
};

class dse{
    public:
        dse(double d, int s, int e, dse** s_user, dse **e_user);
        double d;
        int s,e;
        dse* parent;
        dse* start;
        dse* end;
        dse** s_user, **e_user;
        bool operator<(const dse o2)const;
        void connect(dse* nb);
        bool is_available();
        void set_used();
        bool is_used;
};
